#!/usr/bin/env python

# Rnmap server v. 0.9
#
# Copyright (C) 2000-2002 Tuomo Makinen (tmakinen@pp.htv.fi)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""
Rnmap server v. 0.9

Copyright (C) 2000-2002 Tuomo Makinen
Redistributable under the terms of the GNU General Public License

Usage:
    -p port         Listen on specified port (default: 3418)
    -n path         Nmap path (default: /usr/bin/nmap)
    -h address      Listen address (default: 0.0.0.0)
    -r dir          Directory to chroot to after initialization
    -u uid          Run rnmap server as a specified uid after initialization
    -g gid          Run rnmap server as a specified gid after initialization
    -c dir          Configuration files directory (default .)
    -l dir          Log file directory (default is same as configuration files)

    --help          Prints this message
    --version       Show version of rnmap server
    --nofilter      Turns filtering rules off
    --noplain       Refuse plaintext connections
    --nocipher      Turns encryption mode off
"""

import socket
import os
import string
import sys
import md5
import time
import getopt
import signal
import base64
import select
import fcntl
from errno import EINTR

if sys.hexversion < 0x20200f0:
    import FCNTL
else:
    FCNTL = fcntl

sys.path[:0] = ['../lib']

from sec import *

## Globals
#----------------------------------------------------------
# Codes
PLAINTXTCON   = "100:"
CIPHERTXTCON  = "101:"
RSAKEYREQ     = "102:"
SENDRSAKEY    = "103:"
INTRSCANREQ   = "104"
SUCCESFUL     = "201:" # Succesful scan
INTRPTDSCAN   = "203:" # Interrupted scan
AUTHDENI      = "401:Access denied"
SCANFAIL      = "402:Operation failed"
UNKNOWNERROR  = "403:Unknown command"
BADCHARS      = "404:Operation denied"
UNSUPPORTED   = "405:No encryption support"
NOPLAINTEXT   = "406:Plaintext connection refused"
TOOWEAK       = "407:Too weak encryption key"

# NOTHESECHARS is tuple which has characters that can cause
# unwanted action, and we don't want pass them to system.
NOTHESECHARS = ("|",
                ";",
                "<",
                ">",
                "&")

CRLF = "\r\n"
RECVBUFFERSIZE = 4096
BACKLOG = 5

Server = ''
Port = 3418
Nmapl = "/usr/bin/nmap"
Filteruse = 1
Banplain = 0
Chroot_stat = 0
Process_uid = None
Process_gid = None
Active_childs = []

# Default RSA key length, change this if you need.
Rsakeylength = 1024

try:
    Conf_dir = os.environ["RNMAPDIR"]
except KeyError:
    Conf_dir = os.getcwd()
#----------------------------------------------------------

#==========================================================================
# read configuration files ...
#==========================================================================
# filter.conf is optional file, if omitted we're not using filtering rules.

def confreader():
    global Filteruse
    global Userlist
    Userlist = {}
    if Filteruse == 1:
        global Fmethod
        try:
            # Try to open filtering configuration
            fp = open(Fconf_file)
            tmplist = fp.readlines()
            fp.close()
            global Clist
            Clist = []
            for i in xrange(len(tmplist)):
                if "#" in tmplist[i]:
                    continue
                elif tmplist[i][0:] == '\012':
                    continue
                elif tmplist[i][0] == "!":
                    Fmethod = tmplist[i][9:16]
                else:
                    Clist.append(tmplist[i])
            for j in xrange(len(Clist)):
                Clist[j] = Clist[j][:-1]
        except:
            # No filtering rules used
            Filteruse = None
            Fmethod = None
    # userlist file
    try:
        fp = open(Ulist_file)
        ul = fp.readlines()
        fp.close()
        for i in xrange(len(ul)):
            stopsign = 0
            for j in xrange(len(ul[i])): 
                if ul[i][0] == "#":
                    continue
                elif ul[i][0] == " ":
                    continue
                elif ul[i][j] == ":" and j != 1:
                    if stopsign == 0:
                        Userlist[ul[i][:j]] = base64.decodestring(ul[i][j+1:])
                        stopsign = 1
    except:
        # Can't load users.list ...
        conf_stat = -1
    else:
        conf_stat = None
    return conf_stat

#==========================================================================

def waitconnection():
    try:
        sockobj = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sockobj.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sockobj.bind((Server, Port))
        sockobj.listen(BACKLOG)
    except socket.error, e:
        fatalerror(e[1])
    # --
    log_startup()
    while 1:
        try:
            rlist = select.select([sockobj], [], [], 5.0)[0]
        except select.error, e:
            if e[0] == EINTR:
                continue
        else:
            if sockobj in rlist:
                try:
                    conn, addr = sockobj.accept()
                except socket.error, e:
                    pass
                else:
                    reap_childrens()
                    try:
                        pid = os.fork()
                    except OSError:
                        conn.close()
                        continue
                    if pid == 0:
                        # Close listening socket
                        sockobj.close()
                        # Set default signal handling for child
                        signal.signal(signal.SIGTERM, signal.SIG_DFL)
                        signal.signal(signal.SIGHUP, signal.SIG_DFL)
                        # --
                        newconnection(conn, addr)
                        os._exit(0)
                    else:
                        Active_childs.append(pid)
                        # Close connected socket
                        conn.close()
            else:
                reap_childrens()

def reap_childrens():
    while Active_childs:
        try:
            pid, status = os.waitpid(0, os.WNOHANG)
        except OSError:
            pid = None
        if not pid:
            break
        Active_childs.remove(pid)
            
def newconnection(conn, addr):
    if Filteruse != None and filter(addr[0]) == 0:
        message = AUTHDENI
    else:
        try:
            data = conn.recv(RECVBUFFERSIZE)
        except socket.error:
            data = None
        if not data:
            # Connection is probably broken
            conn.close()
            return
        else:
            if data[:4] == PLAINTXTCON and Banplain == 1:
                logwriter("no_plain", addr[0])
                message = NOPLAINTEXT
            elif Use_cipher == 0 and data[:4] in (CIPHERTXTCON, RSAKEYREQ):
                logwriter("no_cipher", addr[0])
                message = UNSUPPORTED
            elif data[:4] == RSAKEYREQ and Use_cipher == 1:
                message = send_pubkey(data, addr[0])
            else:
                message = get_message_data(data, addr[0], conn)

    totalsent = 0
    while totalsent < len(message):
        try:
            sent = conn.send(message[totalsent:])
        except socket.error:
            break
        totalsent = totalsent + sent
    conn.close()
            
def get_message_data(message, host, conn):
    if Use_cipher == 1 and message[:4] == CIPHERTXTCON:
        com_mode = 1
        iv = message[-8:]
        akey_start = string.find(message, ":: ", 4)
        akey = message[akey_start+3:-8]
        try:
            key = RSA_key.decrypt(akey)
        except:
            key = None
        else:
            key = key[5:]
            if len(key) < 16:
                return TOOWEAK
            else:
                message = decipher(message[4:akey_start], key, iv)
        if key == None or message == None:
            info = UNKNOWNERROR
        else:
            info = process_message(message, host, com_mode, conn, key) 
    elif message[:4] == PLAINTXTCON:
        com_mode = 0
        message = message[4:]
        info = process_message(message, host, com_mode, conn)
    else:
        info = UNKNOWNERROR
    return info

def process_message(message, host, com_mode, conn, key=None):
    message = cleaner(message)
    usr_point = string.find(message, ":")
    login = message[:usr_point]
    cmd_point = string.rfind(message, ":")
    cmd = message[cmd_point+1:]
    passwd = md5.new(message[usr_point+1:cmd_point]).digest()
    cmd = string.replace(cmd, "###", ":")
    if com_mode == 1:
        logwriter("secure_connection", host, login)
    elif com_mode == 0:
        logwriter("plain_connection", host, login)
    if (cmd_point != -1 and usr_point != -1) and (cmd_point != usr_point):
        auth = authenticate(login, passwd)
        if auth == "true":
            scanstat = None
            for i in xrange(len(cmd)): # Check chars 
                if cmd[i] in NOTHESECHARS:
                    scanstat = 0
                    break
                else:
                    scanstat = 1
            if scanstat == 0:
                logwriter("illegal_parameters", host, login, cmd)
                info = BADCHARS   
            elif scanstat == 1:
                info = start_scanner(login, host, cmd, com_mode, conn, key)
            else:
                info = SCANFAIL
        elif auth == "false":
            logwriter("bad_password", host, login)
            info = AUTHDENI
        elif auth == "nouser":
            logwriter("unknown_user", host, login)
            info = AUTHDENI
    else:
        info = UNKNOWNERROR
    if Use_cipher == 1 and com_mode == 1:
        iv = key_generator(8)
        info = encipher(info, key, iv)
        info = info + iv
    return info
    
def start_scanner(login, host, cmd, com_mode, conn, key=None):
    outchunk = ''
    scan_result = ''
    data = ''
    connected = 1
    args = [Nmapl,] + string.split(cmd + ' - ')
    # --
    pRc, cWp = os.pipe()
    child_pid = os.fork()
    if child_pid == 0:
        os.dup2(cWp, 1)     # Combine stderr and stdout
        os.dup2(cWp, 2)     ###########################
        # Ensure that stray fd's are closed
        for i in xrange(3, 256):
            try:
                os.close(i)
            except:
                pass
        try:
            os.execvp(Nmapl, args)
        except:
            os._exit(1)
    os.close(cWp)
    fromchild = os.fdopen(pRc, 'r')
    # --
    fcntl.fcntl(fromchild.fileno(), FCNTL.F_SETFL, os.O_NONBLOCK)
    outfd = fromchild.fileno()
    connfd = conn.fileno()
    
    while 1:
        rlist = select.select([outfd, connfd], [], [])[0]
        if outfd in rlist:
            outchunk = fromchild.read()
            if not outchunk:
                break
            else:
                scan_result = scan_result + outchunk
        if connfd in rlist:
            try:
                data = conn.recv(RECVBUFFERSIZE)
            except socket.error:
                data = None
            if not data:
                connected = 0
            else:
                if Use_cipher == 1 and com_mode == 1:
                    iv = data[-8:]
                    data = decipher(data[:-8], key, iv)
                data = cleaner(data)
            if not connected or data == INTRSCANREQ:
                # Send SIGTERM to process
                try:
                    os.kill(child_pid, signal.SIGTERM)
                except OSError:
                    pass
        time.sleep(0.1)

    # Close pRc and reap children
    os.close(pRc)
    try:
        pid, status = os.waitpid(child_pid, 0)
    except OSError:
        pass
    
    if not connected:
        # Client has closed connection
        conn.close()
        os._exit(0)
    else:
        if data == INTRSCANREQ:
            logwriter("interrupted_scan", host, login, cmd)
            result = INTRPTDSCAN + scan_result
        else:
            logwriter("succesful_scan", host, login, cmd)
            result = SUCCESFUL + scan_result
        return result

def authenticate(login, passwd):
    if Userlist.has_key(login):
        if Userlist[login] == passwd:
            authentication = "true"
        else:
            authentication = "false"
    else:
        authentication = "nouser"
    return (authentication)

def send_pubkey(message, host):
    message = cleaner(message)
    if Userlist.has_key(message[4:]):
        sharedsec = Userlist[message[4:]]
        iv = key_generator(8)
        info = encipher(SENDRSAKEY + Publickey, sharedsec, iv)
        info = info + iv
        logwriter("asymk_request", host, message[4:])
    else:
        info = AUTHDENI
    return info

def logwriter(information, param1=None, param2=None, cmd=None):
    pid = os.getpid()
    if information == "secure_connection":
        wlog = "[" + time.ctime(time.time()) + "][%d] user %s from %s connected with encryption mode.\n"\
               % (pid, param2, param1)
    elif information == "plain_connection":
        wlog = "[" + time.ctime(time.time()) + "][%d] user %s from %s connected with plaintext mode.\n"\
               % (pid, param2, param1)
    elif information == "succesful_scan":
        wlog = "[" + time.ctime(time.time()) + "][%d] user %s from %s performed scan: %s\n"\
               % (pid, param2, param1, cmd)
    elif information ==  "bad_password":
        wlog = "[" + time.ctime(time.time()) + "][%d] user %s from %s tried to connect with wrong password.\n"\
               % (pid, param2, param1)
    elif information == "asymk_request":
        wlog = "[" + time.ctime(time.time()) + "][%d] user %s from %s requested public key.\n"\
               % (pid, param2, param1)
    elif information == "no_plain":
        wlog = "[" + time.ctime(time.time()) + "][%d] host %s connected with unsupported plaintext mode.\n"\
               % (pid, param1)
    elif information == "no_cipher":
        wlog = "[" + time.ctime(time.time()) + "][%d] host %s connected with unsupported ciphertext mode.\n"\
               % (pid, param1)
    elif information == "unknown_user":
        wlog = "[" + time.ctime(time.time()) + "][%d] illegal user %s from %s tried to connect.\n"\
               % (pid, param2, param1)
    elif information == "failed_scan":
        wlog = "[" + time.ctime(time.time()) + "][%d] user %s from %s performed failed scan: %s\n"\
               % (pid, param2, param1, cmd)
    elif information == "banned_host":
        wlog = "[" + time.ctime(time.time()) + "][%d] banned host: %s dropped.\n"\
               % (pid, param1)
    elif information ==  "illegal_parameters":
        wlog = "[" + time.ctime(time.time()) + "][%d] user %s from %s used illegal parameters: %s\n"\
               % (pid, param2, param1, cmd)
    elif information == "interrupted_scan":
        wlog = "[" + time.ctime(time.time()) + "][%d] user %s from %s interrupted scan: %s\n"\
               % (pid, param2, param1, cmd)
    elif information == "startup":
        if Use_cipher == 0:
            wlog = "[" + time.ctime(time.time()) + "][%d] rnmap server started.\n"\
                   % (param1)
        elif Use_cipher == 1:
            wlog = "[" + time.ctime(time.time()) + "][%d] rnmap server started with encryption support.\n"\
                   % (param1)    
        wlog_listen = "[" + time.ctime(time.time()) + "][%d] rnmap server listening on port %d\n"\
                % (pid, param2)
        wlog = wlog + wlog_listen
    elif information == "sigterm":
        wlog = "[" + time.ctime(time.time()) + "][%d] received the SIGTERM signal.\n"\
               % (pid)
    elif information == "sighup":
        wlog = "[" + time.ctime(time.time()) + "][%d] received the SIGHUP signal.\n"\
               % (pid)
    elif information == "ulist_failed":
        wlog = "[" + time.ctime(time.time()) + "][%d] error reading users.list.\n"\
               % (pid)
    elif information == "chrooted":
        wlog = "[" + time.ctime(time.time()) + "][%d] rnmap server chrooted to %s\n" \
               % (pid, Chrootdir)
    elif information == "refuse_plain":
        wlog = "[" + time.ctime(time.time()) + "][%d] refusing plaintext connections.\n" \
               % (pid)
    elif information == "failed_startup":
        wlog = "[" + time.ctime(time.time()) + "][%d] %s" % (pid, param1)
    #---    
    try:
        fp = open(Rnmapdlog, "a")
        fp.write(wlog)
        fp.close()
    except IOError, e:
        pass
        
def cleaner(message):
    if len(message) >= 2:
        if message[-2:] == CRLF:
            message = message[:-2]
        elif message[-1:] in CRLF:
            message = message[:-1]
    return message

def filter(host):
    check = None
    hcheck = host
    if Fmethod == "drpexcp":
        if hcheck in Clist:
            check = 1
        else:
            check = 0
    if Fmethod == "allexcp":
        if hcheck in Clist:
            check = 0
        else:
            check = 1
    return check

def signal_handler_sighup(signum, frame):
    logwriter("sighup")
    conf_stat = confreader()
    if conf_stat != None:
        logwriter(conf_stat)

def signal_handler_sigterm(signum, frame):
    logwriter("sigterm")
    try:
        os.unlink(Rnmapdpid)
    finally:
        sys.exit()

def fatalerror(error):
    logwriter("failed_startup", error + " - server startup failed.\n")
    sys.exit(1)
    
class nulldevice:
    def write(self, s):
        pass
    
def chroot_process():
    try:
        os.chroot(Chrootdir)
    except AttributeError:
        fatalerror("can't chroot() (you need python 2.2 or newer for it)")
    except OSError, e:
        fatalerror("can't chroot() to " + e.filename + ": " + e.strerror)
        
def become_daemon():
    try:
        pid = os.fork()
    except (OSError, AttributeError):
        fatalerror("Can't fork()")
    if pid != 0:
        sys.exit()
    else:
        os.setsid()
        if Chroot_stat == 1:
            chroot_process()
        os.chdir("/")
        sys.stdin.close()
        sys.stdout = nulldevice()
        sys.stderr = nulldevice()
        # Ensure that stray fd's are closed
        for i in range(3, 256):
            try:
                os.close(i)
            except:
                pass

def log_startup():
    # Write pid to disk
    try:
        fp = open(Rnmapdpid, "w")
    except IOError:
        fatalerror("can't write pid file to disk")
    else:
        run = os.getpid()
        fp.write(`run` + "\012")
        fp.close()
    # --
    logwriter("startup", Pid, Port)
    if Chroot_stat == 1:
        logwriter("chrooted")
    if Use_cipher == 1 and Banplain == 1:
        logwriter("refuse_plain")
    if Stat == -1:
        logwriter("ulist_failed")
        
# Fetch command line parameters.
try:
    opts, args = getopt.getopt(sys.argv[1:], 'p:n:h:r:c:l:u:g:', ['help', 'nofilter', 'version', 'noplain', 'nocipher'])
except:
    sys.stderr.write("Try 'rnmap_server.py --help' for more information.\n")
    sys.exit(1)

for o, a in opts:
    if o == '-h': Server = a
    if o == '-p':
        Port = a
    if o == '-n': Nmapl = a
    if o == '-r':
        Chroot_stat = 1
        Chrootdir = a
    if o == '-u':
        Process_uid = a
    if o == '-g':
        Process_gid = a
    if o == '-c':
        Conf_dir = a
    if o == '-l':
        Log_dir = a
    if o == '--help':
        print __doc__
        sys.exit()
    if o == '--version':
        print "Rnmap server v. 0.9"
        sys.exit()
    if o == '--nofilter':
        Filteruse = None
        Fmethod = None
    if o == '--noplain':
        Banplain = 1
    if o == '--nocipher':
        Use_cipher = 0

if Chroot_stat == 1:
    if Conf_dir == os.getcwd():
        Conf_dir = './'
try:
    if Log_dir:
        pass
except NameError:
    Log_dir = Conf_dir

# Set default umask
os.umask(0066)

if Use_cipher == 1:
    import pubkey
    try:
        RSA_key, Publickey, pubkey_stat = pubkey.pub_keys(Rsakeylength, Conf_dir)
    except:
        pubkey_stat = 0
        
# Few globals
Fconf_file = os.path.join(Conf_dir, "filter.conf")
Ulist_file = os.path.join(Conf_dir, "users.list")
Rnmapdpid = os.path.join(Conf_dir, "rnmapd.pid")
Rnmapdlog = os.path.join(Log_dir, "rnmap_server.log")

# Read configuration files and getpid()
Stat = confreader()
Pid = os.getpid()

# Daemonize
become_daemon()

# Set signal handlers
signal.signal(signal.SIGTERM, signal_handler_sigterm)
signal.signal(signal.SIGHUP, signal_handler_sighup)

### Make final checking ... -----------------------------------------------
if Process_gid != None:
    try:
        os.setgid(string.atoi(Process_gid))
    except:
        fatalerror("invalid gid specification")
if Process_uid != None:
    try:
        os.setuid(string.atoi(Process_uid))
    except:
        fatalerror("invalid uid specification")
        
if type(Port) == type(''):
    try:
        Port = string.atoi(Port)
    except ValueError:
        fatalerror("invalid port specification")
        
if os.access(Nmapl, os.F_OK) == 0:
    fatalerror(Nmapl + " doesn't exists! Specify Nmap path with -n option")
    
if Use_cipher == 0 and Banplain == 1:
    fatalerror("server has no encryption support")
    
if Use_cipher == 1:
    if check_randomsrc() == 0:
        fatalerror("no random source (/dev/urandom)")
    if pubkey_stat == 0:
        fatalerror("error reading RSA keys")

# -------------------------------------------------------------------------

# Start work
waitconnection()
